@@ -28,7 +28,14 @@
 
 #include "iso9660.h"
 #include <ctype.h>
+
+#ifndef VMS
 #include <time.h>
+#else
+#include <sys/time.h>
+#include "vms.h"
+#endif
+
 #include <stdlib.h>
 #include <sys/stat.h>
 
@@ -40,7 +47,7 @@
 
 struct directory * root = NULL;
 
-static char version_string[] = "mkisofs v1.00";
+static char version_string[] = "mkisofs v1.01";
 
 FILE * discimage;
 unsigned int next_extent = 0;
@@ -49,18 +56,25 @@ unsigned int path_table_size = 0;
 unsigned int path_table[4] = {0,};
 unsigned int path_blocks = 0;
 struct iso_directory_record root_record;
-int use_RockRidge = 0;
-int verbose = 0;
-int all_files  = 0;
-int follow_links = 0;
-int generate_tables = 0;
 static int timezone_offset;
 char * extension_record = NULL;
 int extension_record_extent = 0;
 static  int extension_record_size = 0;
 
+/* These variables are associated with command line options */
+int use_RockRidge = 0;
+int verbose = 0;
+int all_files  = 0;
+int follow_links = 0;
+int generate_tables = 0;
 char * preparer = NULL;
 char * publisher = NULL;
+int omit_period = 0;             /* Violates iso9660, but these are a pain */
+int transparent_compression = 0; /* So far only works with linux */
+int omit_version_number = 0;     /* May violate iso9660, but noone uses vers*/
+int RR_relocation_depth = 6;     /* Violates iso9660, but most systems work */
+int full_iso9660_filenames = 0;  /* Used with Amiga.  Disc will not work with
+				  DOS */
 
 
 char * path_table_l = NULL;
@@ -69,7 +83,7 @@ int goof = 0;
 
 void usage(){
 	fprintf(stderr,"Usage:\n");
-	fprintf(stderr,"mkisofs [-o outfile] [-R] [-V volid] [-v] [-a] [-T] [-p preparer] [-P publisher] [-x path -x path ...] path\n");
+	fprintf(stderr,"mkisofs [-o outfile] [-R] [-V volid] [-v] [-a] [-T]\n [-l] [-d] [-V] [-D] [-p preparer] [-P publisher] [-z] [-x path -x path ...] path\n");
 	exit(1);
 }
 
@@ -113,6 +127,7 @@ int FDECL2(iso9660_date,char *, result, time_t, ctime){
 int FDECL3(iso9660_file_length,const char*, name, struct directory_entry *, sresult, 
 			int, dirflag){
   int seen_dot = 0;
+  int seen_semic = 0;
   char * result;
   int priority = 32767;
   int tildes = 0;
@@ -136,22 +151,41 @@ int FDECL3(iso9660_file_length,const char*, name, struct directory_entry *, sres
 
   pnt = name;
   while(*pnt){
+#ifdef VMS
+    if(strcmp(pnt,".DIR;1") == 0) break;
+#endif
     if(*pnt == '#') {priority = 1; pnt++; continue; };
     if(*pnt == '~') {priority = 1; tildes++; pnt++; continue;};
+    if(*pnt == ';') {seen_semic = 1; *result++ = *pnt++; continue; };
     if(ignore) {pnt++; continue;};
-    if(*pnt == '.') {
-      if (seen_dot) {ignore++; continue;}
-      if(result) *result++ = '.';
-      seen_dot++;
-    } else if (seen_dot) {
-      if(chars_after_dot < 3) {
-	chars_after_dot++;
-	if(result) *result++ = toupper(*pnt);
+    if(seen_semic){
+      if(*pnt >= '0' && *pnt <= '9') *result++ = *pnt;
+      extra++;
+      pnt++;
+      continue;
+    };
+    if(full_iso9660_filenames) {
+      /* Here we allow a more relaxed syntax. */
+      if(*pnt == '.') {
+	if (seen_dot) {ignore++; continue;}
+	seen_dot++;
       }
-    } else {
-      if(chars_before_dot < 8) {
-	chars_before_dot++;
-	if(result) *result++ = toupper(*pnt);
+      if(current_length < 30) *result++ = (islower(*pnt) ? toupper(*pnt) : *pnt);
+    } else { /* Dos style filenames */
+      if(*pnt == '.') {
+	if (seen_dot) {ignore++; continue;}
+	if(result) *result++ = '.';
+	seen_dot++;
+      } else if (seen_dot) {
+	if(chars_after_dot < 3) {
+	  chars_after_dot++;
+	  if(result) *result++ = (islower(*pnt) ? toupper(*pnt) : *pnt);
+	}
+      } else {
+	if(chars_before_dot < 8) {
+	  chars_before_dot++;
+	  if(result) *result++ = (islower(*pnt) ? toupper(*pnt) : *pnt);
+	};
       };
     };
     current_length++;
@@ -171,15 +205,17 @@ int FDECL3(iso9660_file_length,const char*, name, struct directory_entry *, sres
   };
     
   if (!dirflag){
-    if (!seen_dot) {
+    if (!seen_dot && !omit_period) {
       if (result) *result++ = '.'; 
       extra++;
     };
-    if(result){
-      *result++ = ';';
-      *result++ = '1';
-    };
-    extra += 2;
+    if(!omit_version_number && !seen_semic) {
+      if(result){
+	*result++ = ';';
+	*result++ = '1';
+      };
+      extra += 2;
+    }
   };
 		    
   if(result) *result++ = 0;
@@ -200,7 +236,7 @@ int FDECL2(main, int, argc, char **, argv){
 
   outfile = NULL;
   volid = "CDROM";
-  while ((c = getopt(argc, argv, "o:V:RfvaTp:P:x:")) != EOF)
+  while ((c = getopt(argc, argv, "o:V:RfvaTp:P:x:dDlNz")) != EOF)
     switch (c)
       {
       case 'p':
@@ -217,6 +253,18 @@ int FDECL2(main, int, argc, char **, argv){
 		exit(1);
 	};
 	break;
+      case 'd':
+	omit_period++;
+	break;
+      case 'D':
+	RR_relocation_depth = 32767;
+	break;
+      case 'l':
+	full_iso9660_filenames++;
+	break;
+      case 'N':
+	omit_version_number++;
+	break;
       case 'o':
 	outfile = optarg;
 	break;
@@ -238,6 +286,14 @@ int FDECL2(main, int, argc, char **, argv){
       case 'T':
 	generate_tables++;
 	break;
+      case 'z':
+#ifdef VMS
+	fprintf(stderr,"Transparent compression not supported with VMS\n");
+	exit(1);
+#else
+	transparent_compression++;
+#endif
+	break;
       case 'x':
         exclude(optarg);
 	break;
@@ -262,11 +318,13 @@ int FDECL2(main, int, argc, char **, argv){
 	  exit(1);
   };
 
+#ifndef VMS
   if(scan_tree[strlen(scan_tree)-1] != '/') {
     scan_tree = (char *) malloc(strlen(argv[optind])+2);
     strcpy(scan_tree, argv[optind]);
     strcat(scan_tree, "/");
   };
+#endif
 
   if(use_RockRidge){
 	extension_record = generate_rr_extension_record("RRIP_1991A",
@@ -275,7 +333,7 @@ int FDECL2(main, int, argc, char **, argv){
   };
 
   stat(argv[optind], &statbuf);
-  add_directory_hash(statbuf.st_dev, statbuf.st_ino);
+  add_directory_hash(statbuf.st_dev, STAT_INODE(statbuf));
 
   de.filedir = root;  /* We need this to bootstrap */
   scan_directory_tree(argv[optind], &de);
@@ -338,6 +396,10 @@ int FDECL2(main, int, argc, char **, argv){
   fprintf(stderr,"Max brk space used %x\n", 
 	  ((unsigned int)sbrk(0)) - mem_start);
   fprintf(stderr,"%d extents written (%d Mb)\n", last_extent, last_extent >> 9);
+#ifdef VMS
+  return 1;
+#else
   return 0;
+#endif
 }
 
