@@ -27,15 +27,24 @@
 
 #ifndef VMS
 #include <unistd.h>
+#if	!defined(__QNX__) && !defined(__NetBSD__)
 #include <sys/sysmacros.h>
 #endif
+#else
+#include <sys/file.h>
+#include <vms/fabdef.h>
+#include "vms.h"
+extern char * strdup(const char *);
+#endif
 
 #include "mkisofs.h"
 #include "iso9660.h"
 
 #include <sys/stat.h>
 
-#ifdef VMS
+#include "exclude.h"
+
+#ifdef NON_UNIXFS
 #define S_ISLNK(m)	(0)
 #define S_ISSOCK(m)	(0)
 #define S_ISFIFO(m)	(0)
@@ -72,7 +81,7 @@ void FDECL1(sort_n_finish, struct directory *, this_dir)
   int new_reclen;
   char *  c;
   int tablesize = 0;
-  char newname[15],  rootname[15];
+  char newname[34],  rootname[34];
 
   /* Here we can take the opportunity to toss duplicate entries from the
      directory.  */
@@ -107,12 +116,27 @@ void FDECL1(sort_n_finish, struct directory *, this_dir)
 	  /* OK, handle the conflicts.  Try substitute names until we come
 	     up with a winner */
 	  strcpy(rootname, s_entry->isorec.name);
+	  if(full_iso9660_filenames) {
+	    if(strlen(rootname) > 27) rootname[27] = 0;
+	  }
 	  c  = strchr(rootname, '.');
 	  if (c) *c = 0;
 	  count = 0;
 	  while(count < 1000){
 		  sprintf(newname,"%s.%3.3d%s", rootname,  count,
-			  (s_entry->isorec.flags[0] == 2 ? "" : ";1"));
+			  (s_entry->isorec.flags[0] == 2 || 
+			   omit_version_number ? "" : ";1"));
+
+#ifdef VMS
+		  /* Sigh.  VAXCRTL seems to be broken here */
+		  { int ijk = 0;
+		    while(newname[ijk]) {
+		      if(newname[ijk] == ' ') newname[ijk] = '0';
+		      ijk++;
+		    };
+		  }
+#endif
+
 		  if(!find_file_hash(newname)) break;
 		  count++;
 	  };
@@ -125,7 +149,7 @@ void FDECL1(sort_n_finish, struct directory *, this_dir)
 	     of these two beasts should get the name changed */
 	  
 	  if(s_entry->priority < s_entry1->priority) {
-		  fprintf(stderr,"Using %s for  %s/%s (%s)\n", newname,  this_dir->whole_name, s_entry->name, s_entry1->name);
+		  fprintf(stderr,"Using %s for  %s%s%s (%s)\n", newname,  this_dir->whole_name, SPATH_SEPARATOR, s_entry->name, s_entry1->name);
 		  s_entry->isorec.name_len[0] =  strlen(newname);
 		  new_reclen =  sizeof(struct iso_directory_record) -
 			  sizeof(s_entry->isorec.name) +
@@ -139,7 +163,7 @@ void FDECL1(sort_n_finish, struct directory *, this_dir)
 		  strcpy(s_entry->isorec.name, newname);
 	  } else {
 		  delete_file_hash(s_entry1);
-		  fprintf(stderr,"Using %s for  %s/%s (%s)\n", newname,  this_dir->whole_name, s_entry1->name, s_entry->name);
+		  fprintf(stderr,"Using %s for  %s%s%s (%s)\n", newname,  this_dir->whole_name, SPATH_SEPARATOR, s_entry1->name, s_entry->name);
 		  s_entry1->isorec.name_len[0] =  strlen(newname);
 		  new_reclen =  sizeof(struct iso_directory_record) -
 			  sizeof(s_entry1->isorec.name) +
@@ -157,7 +181,7 @@ void FDECL1(sort_n_finish, struct directory *, this_dir)
 	  s_entry = s_entry->next;
   };
 
-  if(generate_tables && !find_file_hash("TRANS.TBL;1") && (reloc_dir != this_dir)){
+  if(generate_tables && !find_file_hash("TRANS.TBL") && (reloc_dir != this_dir)){
 	  /* First we need to figure out how big this table is */
 	  for (s_entry = this_dir->contents; s_entry; s_entry = s_entry->next){
 		  if(strcmp(s_entry->name, ".") == 0  ||
@@ -176,7 +200,7 @@ void FDECL1(sort_n_finish, struct directory *, this_dir)
 	table->priority  = 32768;
 	iso9660_date(table->isorec.date, current_time);
 	table->inode = TABLE_INODE;
-	table->dev = UNCACHED_DEVICE;
+	table->dev = (dev_t) UNCACHED_DEVICE;
 	set_723(table->isorec.volume_sequence_number, 1);
 	set_733(table->isorec.size, tablesize);
 	table->size = tablesize;
@@ -184,7 +208,7 @@ void FDECL1(sort_n_finish, struct directory *, this_dir)
 	table->name = strdup("<translation table>");
 	table->table = (char *) malloc(ROUND_UP(tablesize));
 	memset(table->table, 0, ROUND_UP(tablesize));
-	iso9660_file_length  ("TRANS.TBL;1", table, 1);
+	iso9660_file_length  ("TRANS.TBL", table, 1);
 		
 	if(use_RockRidge){
 		fstatbuf.st_mode = 0444 | S_IFREG;
@@ -260,12 +284,43 @@ void FDECL1(sort_n_finish, struct directory *, this_dir)
      Do not split a directory entry across a sector boundary */
   
   s_entry = this_dir->contents;
+  this_dir->ce_bytes = 0;
   while(s_entry){
 	  new_reclen = s_entry->isorec.length[0];
 	  if ((this_dir->size & (SECTOR_SIZE - 1)) + new_reclen >= SECTOR_SIZE)
 		  this_dir->size = (this_dir->size + (SECTOR_SIZE - 1)) & 
 			  ~(SECTOR_SIZE - 1);
 	  this_dir->size += new_reclen;
+
+	  /* See if continuation entries were used on disc */
+	  if(use_RockRidge && 
+	     s_entry->rr_attr_size != s_entry->total_rr_attr_size) {
+	    unsigned char * pnt;
+	    int len;
+	    int nbytes;
+
+	    pnt = s_entry->rr_attributes;
+	    len = s_entry->total_rr_attr_size;
+
+	    /* We make sure that each continuation entry record is not
+	       split across sectors, but each file could in theory have more
+	       than one CE, so we scan through and figure out what we need. */
+
+	    while(len > 3){
+	      if(pnt[0] == 'C' && pnt[1] == 'E') {
+		nbytes = get_733(pnt+20);
+		
+		if((this_dir->ce_bytes & (SECTOR_SIZE - 1)) + nbytes >=
+		   SECTOR_SIZE) this_dir->ce_bytes = 
+		     ROUND_UP(this_dir->ce_bytes);
+		/* Now store the block in the ce buffer */
+		this_dir->ce_bytes += nbytes;
+		if(this_dir->ce_bytes & 1) this_dir->ce_bytes++;
+	      };
+	      len -= pnt[2];
+	      pnt += pnt[2];
+	    }
+	  }
 	  s_entry = s_entry->next;
   }
 }
@@ -305,7 +360,7 @@ static void generate_reloc_directory()
 	root->contents->priority  = 32768;
 	iso9660_date(root->contents->isorec.date, current_time);
 	root->contents->inode = UNCACHED_INODE;
-	root->contents->dev = UNCACHED_DEVICE;
+	root->contents->dev = (dev_t) UNCACHED_DEVICE;
 	set_723(root->contents->isorec.volume_sequence_number, 1);
 	iso9660_file_length (reloc_dir->de_name, root->contents, 1);
 
@@ -360,7 +415,7 @@ static void FDECL1(increment_nlink, struct directory_entry *, s_entry){
   int len, nlink;
 
   pnt = s_entry->rr_attributes;
-  len = s_entry->rr_attr_size;
+  len = s_entry->total_rr_attr_size;
   while(len){
     if(pnt[0] == 'P' && pnt[1] == 'X') {
       nlink =  get_733(pnt+12);
@@ -391,13 +446,13 @@ void finish_cl_pl_entries(){
 
 	  /* First fix the PL pointer in the directory in the rr_reloc dir */
 	  s_entry1 = d_entry->contents->next;
-	  set_733(s_entry1->rr_attributes +  s_entry1->rr_attr_size - 8,
+	  set_733(s_entry1->rr_attributes +  s_entry1->total_rr_attr_size - 8,
 		  s_entry->filedir->extent);
 
 	  /* Now fix the CL pointer */
 	  s_entry1 = s_entry->parent_rec;
 
-	  set_733(s_entry1->rr_attributes +  s_entry1->rr_attr_size - 8,
+	  set_733(s_entry1->rr_attributes +  s_entry1->total_rr_attr_size - 8,
 		  d_entry->extent);
 
 	  s_entry->filedir = reloc_dir;  /* Now we can fix this */
@@ -419,7 +474,7 @@ void finish_cl_pl_entries(){
  * directory entries, so that we can determine how large each directory is.
  */
 
-void 
+int
 FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
   DIR * current_dir;
   char whole_path[1024];
@@ -432,10 +487,21 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
   int new_reclen;
   int deep_flag;
 
+  current_dir = opendir(path);
+  if(!current_dir) {
+	  fprintf(stderr,"Unable to open directory %s\n", path);
+	  de->isorec.flags[0] &= ~2; /* Mark as not a directory */
+	  return 0;
+  };
+
   parent = de->filedir;
   /* Set up the struct for the current directory, and insert it into the
      tree */
 
+#ifdef VMS
+  vms_path_fixup(path);
+#endif
+
   this_dir = (struct directory *) malloc(sizeof(struct directory));
   this_dir->next = NULL;
   new_reclen = 0;
@@ -443,7 +509,7 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
   this_dir->self = de;
   this_dir->contents = NULL;
   this_dir->whole_name = strdup(path);
-  cpnt = strrchr(path, '/');
+  cpnt = strrchr(path, PATH_SEPARATOR);
   if(cpnt)
     cpnt++;
   else
@@ -473,7 +539,7 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
 #ifdef DEBUG
 	  fprintf(stderr,"%s(%d) ", path, this_dir->depth);
 #endif
-	  if(parent->depth >  6) {
+	  if(parent->depth >  RR_relocation_depth) {
 		  fprintf(stderr,"Directories too deep  %s\n", path);
 		  exit(1);
 	  };
@@ -492,11 +558,6 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
 
 /* Now we scan the directory itself, and look at what is inside of it. */
 
-  current_dir = opendir(path);
-  if(!current_dir) {
-	  goof++;
-	  fprintf(stderr,"Fatal goof - %s is not a directory\n", path);
-  };
   while(1==1){
 
     d_entry = readdir(current_dir);
@@ -517,8 +578,19 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
 
     /* Generate the complete ASCII path for this file */
     strcpy(whole_path, path);
-    if(whole_path[strlen(whole_path)-1] != '/') strcat(whole_path, "/");
+#ifndef VMS
+    if(whole_path[strlen(whole_path)-1] != '/')
+      strcat(whole_path, "/");
+#endif
     strcat(whole_path, d_entry->d_name);
+
+    /* Should we exclude this file? */
+    if (is_excluded(whole_path)) {
+      if (verbose) {
+        fprintf(stderr, "Excluded: %s\n",whole_path);
+      }
+      continue;
+    }
 #if 0
     if (verbose)  fprintf(stderr, "%s\n",whole_path);
 #endif
@@ -546,8 +618,8 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
 		    if(use_RockRidge){
 			    status = 0;
 			    statbuf.st_size = 0;
-			    statbuf.st_ino = UNCACHED_INODE;
-			    statbuf.st_dev = UNCACHED_DEVICE;
+			    STAT_INODE(statbuf) = UNCACHED_INODE;
+			    statbuf.st_dev = (dev_t) UNCACHED_DEVICE;
 			    statbuf.st_mode = (statbuf.st_mode & ~S_IFMT) | S_IFREG;
 		    } else {
 			    if(follow_links) fprintf(stderr,
@@ -568,20 +640,29 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
 	       	       
 	    else if(strcmp(d_entry->d_name, ".") && 
 	       strcmp(d_entry->d_name, "..")) {
-		    if(find_directory_hash(statbuf.st_dev, statbuf.st_ino)){
+		    if(find_directory_hash(statbuf.st_dev, STAT_INODE(statbuf))){
 			    fprintf(stderr, "Infinite loop detected (%s)\n", whole_path);
 			    if(!use_RockRidge) continue;
 			    statbuf.st_size = 0;
-			    statbuf.st_ino = UNCACHED_INODE;
-			    statbuf.st_dev = UNCACHED_DEVICE;
+			    STAT_INODE(statbuf) = UNCACHED_INODE;
+			    statbuf.st_dev = (dev_t) UNCACHED_DEVICE;
 			    statbuf.st_mode = (statbuf.st_mode & ~S_IFMT) | S_IFREG;
 		    } else {
 			    lstatbuf = statbuf;
-			    add_directory_hash(statbuf.st_dev, statbuf.st_ino);
+			    add_directory_hash(statbuf.st_dev, STAT_INODE(statbuf));
 		    };
 	    };	    
     };
 
+#ifdef VMS
+    if(!S_ISDIR(lstatbuf.st_mode) && (statbuf.st_fab_rfm != FAB$C_FIX && 
+				      statbuf.st_fab_rfm != FAB$C_STMLF)) {
+      fprintf(stderr,"Warning - file %s has an unsupported VMS record"
+	      " format (%d)\n",
+	      whole_path, statbuf.st_fab_rfm);
+    };
+#endif
+
     if(S_ISREG(lstatbuf.st_mode) && (status = access(whole_path, R_OK))){
       fprintf(stderr, "File %s is not readable (errno = %d) - ignoring\n", 
 	      whole_path, errno);
@@ -592,12 +673,12 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
      If we are set up to follow symlinks, then we skip this checking. */
     if(!follow_links && S_ISDIR(lstatbuf.st_mode) && strcmp(d_entry->d_name, ".") && 
        strcmp(d_entry->d_name, "..")) {
-	    if(find_directory_hash(statbuf.st_dev, statbuf.st_ino)) {
+	    if(find_directory_hash(statbuf.st_dev, STAT_INODE(statbuf))) {
 		    fprintf(stderr,"Directory loop - fatal goof (%s %x %d).\n",
-			    whole_path, statbuf.st_dev, statbuf.st_ino);
+			    whole_path, statbuf.st_dev, STAT_INODE(statbuf));
 		    exit(1);
 	    };
-	    add_directory_hash(statbuf.st_dev, statbuf.st_ino);
+	    add_directory_hash(statbuf.st_dev, STAT_INODE(statbuf));
     };
 
     if (!S_ISCHR(lstatbuf.st_mode) && !S_ISBLK(lstatbuf.st_mode) &&
@@ -635,15 +716,16 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
     s_entry->isorec.interleave[0] = 0;
     if(parent && parent ==  reloc_dir && strcmp(d_entry->d_name,  "..") == 0){
 	    s_entry->inode = UNCACHED_INODE;
-	    s_entry->dev = UNCACHED_DEVICE;
+	    s_entry->dev = (dev_t) UNCACHED_DEVICE;
 	    deep_flag  = NEED_PL;
     } else  {
-	    s_entry->inode = statbuf.st_ino;
+	    s_entry->inode = STAT_INODE(statbuf);
 	    s_entry->dev = statbuf.st_dev;
     };
     set_723(s_entry->isorec.volume_sequence_number, 1);
     iso9660_file_length(d_entry->d_name, s_entry, S_ISDIR(statbuf.st_mode));
     s_entry->rr_attr_size = 0;
+    s_entry->total_rr_attr_size = 0;
     s_entry->rr_attributes = NULL;
 
     /* Directories are assigned sizes later on */
@@ -660,7 +742,7 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
       s_entry->isorec.flags[0] = 2;
 
     if (strcmp(d_entry->d_name,".") && strcmp(d_entry->d_name,"..") && 
-	S_ISDIR(statbuf.st_mode) && this_dir->depth >  6){
+	S_ISDIR(statbuf.st_mode) && this_dir->depth >  RR_relocation_depth){
 		  if(!reloc_dir) generate_reloc_directory();
 
 		  s_entry1 = (struct directory_entry *) 
@@ -706,7 +788,7 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
 		    sprintf(buffer,"D\t%s\n",
 			    s_entry->name);
 		    break;
-#ifndef VMS
+#ifndef NON_UNIXFS
 	    case S_IFBLK:
 		    sprintf(buffer,"B\t%s\t%d %d\n",
 			    s_entry->name,
@@ -730,7 +812,7 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
 		    sprintf(buffer,"S\t%s\n",
 			    s_entry->name);
 		    break;
-#endif
+#endif /* NON_UNIXFS */
 	    case S_IFREG:
 	    default:
 		    sprintf(buffer,"F\t%s\n",
@@ -741,8 +823,13 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
     };
     
     if(S_ISDIR(statbuf.st_mode)){
-	    if (strcmp(d_entry->d_name,".") && strcmp(d_entry->d_name,".."))
-		scan_directory_tree(whole_path, s_entry);
+            int dflag;
+	    if (strcmp(d_entry->d_name,".") && strcmp(d_entry->d_name,"..")) {
+		dflag = scan_directory_tree(whole_path, s_entry);
+	    /* If unable to scan directory, mark this as a non-directory */
+	        if(!dflag)
+		  lstatbuf.st_mode = (lstatbuf.st_mode & ~S_IFMT) | S_IFREG;
+	      }
 	  };
 
   if(use_RockRidge && this_dir == root && strcmp(s_entry->name, ".")  == 0)
@@ -759,6 +846,7 @@ FDECL2(scan_directory_tree,char *, path, struct directory_entry *, de){
   }
   closedir(current_dir);
   sort_n_finish(this_dir);
+  return 1;
 }
 
 
