@@ -30,6 +30,10 @@
 extern char * strdup(const char *);
 #endif
 
+#ifdef VMS
+extern char * strdup(const char *);
+#endif
+
 
 /* Max number of sectors we will write at  one time */
 #define NSECT 16
@@ -113,7 +117,7 @@ static struct deferred_write * dw_head = NULL, * dw_tail = NULL;
 
 static struct directory_entry * sort_dir;
 
-static unsigned int last_extent_written  =0;
+unsigned int last_extent_written  =0;
 static struct iso_primary_descriptor vol_desc;
 static path_table_index;
 
@@ -133,6 +137,12 @@ void FDECL1(assign_directory_addresses, struct directory *, node){
     dir_size = (dpnt->size + (SECTOR_SIZE - 1)) >> 11;
 
     last_extent += dir_size;
+
+    /* Leave room for the CE entries for this directory.  Keep them
+       close to the reference directory so that access will be quick. */
+    if(dpnt->ce_bytes)
+      last_extent += ROUND_UP(dpnt->ce_bytes) >> 11;
+
     if(dpnt->subdir) assign_directory_addresses(dpnt->subdir);
     dpnt = dpnt->next;
   };
@@ -175,12 +185,17 @@ static void FDECL1(write_files, FILE *, outfile){
   while(dwpnt){
 	  if(dwpnt->table) {
 		  xfwrite(dwpnt->table,  1, ROUND_UP(dwpnt->size), outfile);
+		  last_extent_written += ROUND_UP(dwpnt->size) / SECTOR_SIZE;
 		  table_size += dwpnt->size;
-		  fprintf(stderr,"Size %d ", dwpnt->size);
+/*		  fprintf(stderr,"Size %d ", dwpnt->size); */
 		  free(dwpnt->table);
 	  } else {
 
+#ifdef VMS
+		  vms_write_one_file(dwpnt->name, dwpnt->size, outfile);
+#else
 		  write_one_file(dwpnt->name, dwpnt->size, outfile);
+#endif
 		  free(dwpnt->name);
 	  };
 
@@ -203,8 +218,22 @@ static void dump_filelist(){
 #endif
 
 int FDECL2(compare_dirs, const struct directory_entry **, r, const struct directory_entry **, l) {
-  return strcmp((*r)->isorec.name, (*l)->isorec.name);;
-  
+  char * rpnt, *lpnt;
+
+  rpnt = (*r)->isorec.name;
+  lpnt = (*l)->isorec.name;
+
+  while(*rpnt && *lpnt) {
+    if(*rpnt == ';' && *lpnt != ';') return -1;
+    if(*rpnt != ';' && *lpnt == ';') return 1;
+    if(*rpnt == ';' && *lpnt == ';') return 0;
+    if(*rpnt < *lpnt) return -1;
+    if(*rpnt > *lpnt) return 1;
+    rpnt++;  lpnt++;
+  }
+  if(*rpnt) return 1;
+  if(*lpnt) return -1;
+  return 0;
 }
 
 void FDECL1(sort_directory, struct directory_entry **, sort_dir){
@@ -279,8 +308,8 @@ static void FDECL1(assign_file_addresses, struct directory *, dpnt){
       s_hash = find_hash(s_entry->dev, s_entry->inode);
       if(s_hash){
         if(verbose)
-	  fprintf(stderr, "Cache hit for %s/%s\n",s_entry->filedir->de_name, 
-		  s_entry->name);
+	  fprintf(stderr, "Cache hit for %s%s%s\n",s_entry->filedir->de_name, 
+		  SPATH_SEPARATOR, s_entry->name);
         set_733(s_entry->isorec.extent, s_hash->starting_block);
         set_733(s_entry->isorec.size, s_hash->size);
         continue;
@@ -345,7 +374,9 @@ static void FDECL1(assign_file_addresses, struct directory *, dpnt){
 	    } else {
 	      dwpnt->table = NULL;
 	      strcpy(whole_path, s_entry->filedir->whole_name);
+#ifndef VMS
 	      if(strlen(whole_path)) strcat(whole_path, "/");
+#endif
 	      strcat(whole_path, s_entry->name);
 	      dwpnt->name = strdup(whole_path);
 	    };
@@ -384,17 +415,33 @@ static void FDECL1(assign_file_addresses, struct directory *, dpnt){
 }
 
 void FDECL2(generate_one_directory, struct directory *, dpnt, FILE *, outfile){
-  unsigned int total_size;
+  unsigned int total_size, ce_size;
   char * directory_buffer;
+  char * ce_buffer;
+  unsigned int ce_address;
   struct directory_entry * s_entry, *s_entry_d;
   int new_reclen;
-  unsigned int dir_index;
+  unsigned int dir_index, ce_index;
 
   total_size = (dpnt->size + (SECTOR_SIZE - 1)) &  ~(SECTOR_SIZE - 1);
   directory_buffer = (char *) malloc(total_size);
   memset(directory_buffer, 0, total_size);
   dir_index = 0;
 
+  ce_size = (dpnt->ce_bytes + (SECTOR_SIZE - 1)) &  ~(SECTOR_SIZE - 1);
+  ce_buffer = NULL;
+
+  if(ce_size) {
+    ce_buffer = (char *) malloc(ce_size);
+    memset(ce_buffer, 0, ce_size);
+    
+    ce_index = 0;
+    
+    /* Absolute byte address of CE entries for this directory */
+    ce_address = last_extent_written + (total_size >> 11);
+    ce_address = ce_address << 11;
+  }
+
   s_entry = dpnt->contents;
   while(s_entry) {
 
@@ -416,11 +463,44 @@ void FDECL2(generate_one_directory, struct directory *, dpnt, FILE *, outfile){
       if(dir_index & 1)
 	directory_buffer[dir_index++] = 0;
 
-      rockridge_size += s_entry->rr_attr_size;
+      /* If the RR attributes were too long, then write the CE records,
+	 as required. */
+      if(s_entry->rr_attr_size != s_entry->total_rr_attr_size) {
+	unsigned char * pnt;
+	int len, nbytes;
+
+	/* Go through the entire record and fix up the CE entries
+	   so that the extent and offset are correct */
+
+	pnt = s_entry->rr_attributes;
+	len = s_entry->total_rr_attr_size;
+	while(len > 3){
+	  if(pnt[0] == 'C' && pnt[1] == 'E') {
+	    nbytes = get_733(pnt+20);
+
+	    if((ce_index & (SECTOR_SIZE - 1)) + nbytes >=
+	       SECTOR_SIZE) ce_index = ROUND_UP(ce_index);
+
+	    set_733(pnt+4, (ce_address + ce_index) >> 11);
+	    set_733(pnt+12, (ce_address + ce_index) & (SECTOR_SIZE - 1));
+	    
+
+	    /* Now store the block in the ce buffer */
+	    memcpy(ce_buffer + ce_index, 
+		   pnt + pnt[2], nbytes);
+	    ce_index += nbytes;
+	    if(ce_index & 1) ce_index++;
+	  };
+	  len -= pnt[2];
+	  pnt += pnt[2];
+	};
+
+      }
+
+      rockridge_size += s_entry->total_rr_attr_size;
       memcpy(directory_buffer + dir_index, s_entry->rr_attributes, 
 	     s_entry->rr_attr_size);
       dir_index += s_entry->rr_attr_size;
-
     };
     if(dir_index & 1)
 	    directory_buffer[dir_index++] = 0;
@@ -437,6 +517,16 @@ void FDECL2(generate_one_directory, struct directory *, dpnt, FILE *, outfile){
   xfwrite(directory_buffer, 1, total_size, outfile);
   last_extent_written += total_size >> 11;
   free(directory_buffer);
+
+  if(ce_size){
+    if(ce_index != dpnt->ce_bytes)
+      fprintf(stderr,"Continuation entry record length mismatch (%d %d).\n",
+	      ce_index, dpnt->ce_bytes);
+    xfwrite(ce_buffer, 1, ce_size, outfile);
+    last_extent_written += ce_size >> 11;
+    free(ce_buffer);
+  }
+
 }
 
 static void FDECL1(build_pathlist, struct directory *, node){
@@ -499,7 +589,7 @@ void generate_path_tables(){
     };
     npnt = dpnt->de_name;
     if(*npnt == 0 || dpnt == root) npnt = ".";  /* So the root comes out OK */
-    npnt1 = strrchr(npnt, '/');
+    npnt1 = strrchr(npnt, PATH_SEPARATOR);
     if(npnt1) npnt = npnt1 + 1;
 	    
     de = dpnt->self;
@@ -650,14 +740,16 @@ int FDECL2(iso_write, FILE *, outfile, char *, volid){
 #endif
   generate_iso9660_directories(root, outfile);
 
-  if(extension_record) xfwrite(extension_record, 1, SECTOR_SIZE, outfile);
-
+  if(extension_record) {
+    xfwrite(extension_record, 1, SECTOR_SIZE, outfile);
+    last_extent_written++;
+  }
 
   /* Now write all of the files that we need. */
-  fprintf(stderr,"Total extents being written = %d\n", last_extent);
+  fprintf(stderr,"Total extents scheduled to be written = %d\n", last_extent);
   write_files(outfile);
 
-
+  fprintf(stderr,"Total extents actually written = %d\n", last_extent_written);
   /* Hard links throw us off here */
   if(should_write != last_extent){
     fprintf(stderr,"Number of extents written not what was predicted.  Please fix.\n");
